
import datetime as dt

from ...lib.date import local_date_to_time
from ...squeal.tables.pipeline import Pipeline
from ...squeal.tables.source import Source
from ...squeal.tables.statistic import StatisticJournal, Statistic


def build_pipeline(log, session, type, factory, date):
    for cls, args, kargs in Pipeline.all(log, session, type):
        log.info('Building %s (%s, %s)' % (cls, args, kargs))
        yield from cls(log).build(session, factory, date, *args, **kargs)


class Displayer:

    def __init__(self, log):
        self._log = log

    def _journal_at_date(self, s, date, name, owner, constraint):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        return s.query(StatisticJournal).join(Statistic, Source). \
            filter(Statistic.name == name,
                   Source.time >= start,
                   Source.time < finish,
                   Statistic.owner == owner,
                   Statistic.constraint == constraint).one_or_none()

    def _journal_at_time(self, s, time, name, owner, constraint):
        return s.query(StatisticJournal).join(Statistic, Source). \
            filter(Statistic.name == name,
                   Source.time == time,
                   Statistic.owner == owner,
                   Statistic.constraint == constraint).one_or_none()




from collections import defaultdict

from pandas import DataFrame
from pygeotile.point import Point
from sqlalchemy.sql.functions import count

from ..command.args import parser, NamespaceWithVariables, NO_OP
from ..lib.log import make_log
from ..squeal.database import Database
from ..squeal.tables.activity import Activity, ActivityJournal
from ..squeal.tables.source import Interval, Source
from ..squeal.tables.statistic import Statistic, StatisticJournal, StatisticMeasure


def extract(data, instance, *attributes):
    for attribute in attributes:
        if hasattr(instance, attribute):
            data[attribute].append(getattr(instance, attribute))
        else:
            data[attribute].append(None)


class Data:

    def __init__(self, log, db):
        self._log = log
        self._session = db.session()

    def activities(self):
        data, ids = defaultdict(list), []
        for activity in self._session.query(Activity).order_by(Activity.name):
            ids.append(activity.id)
            extract(data, activity, 'name', 'description')
            data['count'].append(self._session.query(count(ActivityJournal.id)).
                                 filter(ActivityJournal.activity == activity).scalar())
        return DataFrame(data, index=ids)

    def activity_journals(self, activity, start=None, finish=None):
        data, times = defaultdict(list), []
        q = self._session.query(ActivityJournal).join(Activity).filter(Activity.name == activity)
        if start:
            q = q.filter(ActivityJournal.time >= start)
        if finish:
            q = q.filter(ActivityJournal.time <= finish)
        for journal in q.order_by(ActivityJournal.time).all():
            times.append(journal.time)
            extract(data, journal, 'name', 'fit_file', 'finish')
        return DataFrame(data, index=times)

    def activity_waypoints(self, activity, time):
        journal = self._session.query(ActivityJournal).join(Activity). \
            filter(Activity.name == activity, ActivityJournal.time == time).one()
        frames = []
        for timespan in journal.timespans:
            data, times = defaultdict(list), []
            for waypoint in timespan.waypoints:
                times.append(waypoint.time)
                extract(data, waypoint, 'latitude', 'longitude', 'heart_rate', 'distance', 'speed')
                if waypoint.latitude and waypoint.longitude:
                    p = Point.from_latitude_longitude(waypoint.latitude, waypoint.longitude)
                    x, y = p.meters
                else:
                    x, y = None, None
                data['x'].append(x)
                data['y'].append(y)
            frames.append(DataFrame(data, index=times))
        return frames

    def statistics(self, *statistics):
        statistic_names, statistic_ids = self._collect_statistics(statistics)
        data = defaultdict(list)
        for statistic in self._session.query(Statistic). \
                filter(Statistic.id.in_(statistic_ids)).order_by(Statistic.name):
            extract(data, statistic, 'name', 'description', 'units', 'summary', 'owner', 'constraint')
            data['count'].append(self._session.query(count(StatisticJournal.id)).
                                 filter(StatisticJournal.statistic == statistic).scalar())
        return DataFrame(data)

    def _collect_statistics(self, statistics):
        if not statistics:
            statistics = ['%']
        statistic_ids, statistic_names = set(), set()
        for statistic in statistics:
            for statistic in self._session.query(Statistic).filter(Statistic.name.like(statistic)).all():
                statistic_ids.add(statistic.id)
                statistic_names.add(statistic.name)
        return statistic_names, statistic_ids

    def _build_statistic_journal_query(self, statistic_ids, q,
            start, finish, owner, constraint):
        q = q.filter(Statistic.id.in_(statistic_ids))
        if start:
            q = q.filter(Source.time >= start)
        if finish:
            q = q.filter(Source.time <= finish)
        if owner:
            q = q.filter(Statistic.owner == owner)
        if constraint:
            q = q.filter(Statistic.constraint == constraint)
        return q

    def statistic_journals(self, *statistics, start=None, finish=None, owner=None, constraint=None, schedule=None):
        statistic_names, statistic_ids = self._collect_statistics(statistics)
        q = self._build_statistic_journal_query(
            statistic_ids, self._session.query(StatisticJournal).join(Statistic, Source),
            start, finish, owner, constraint)
        if schedule:
            q = q.join(Interval).filter(Interval.schedule == schedule)
        raw_data = defaultdict(dict)
        for journal in q.all():  # todo - eager load
            raw_data[journal.time][journal.statistic.name] = journal.value
        data, times = defaultdict(list), []
        for time in sorted(raw_data.keys()):
            times.append(time)
            sub_data = raw_data[time]
            for statistic in statistic_names:
                if statistic in sub_data:
                    data[statistic].append(sub_data[statistic])
                else:
                    data[statistic].append(None)
        return DataFrame(data, index=times)

    def statistic_quartiles(self, *statistics, schedule='m', start=None, finish=None, owner=None, constraint=None):
        statistic_names, statistic_ids = self._collect_statistics(statistics)
        q = self._build_statistic_journal_query(
            statistic_ids, self._session.query(StatisticMeasure).join(StatisticJournal, Statistic, Source),
            start, finish, owner, constraint)
        if schedule:
            q = q.join((Interval, StatisticMeasure.source_id == Interval.id)). \
                filter(Interval.schedule == schedule)
        q = q.filter(StatisticMeasure.quartile != None)
        raw_data = defaultdict(lambda: defaultdict(lambda: [0] * 5))
        for measure in q.all():
            raw_data[measure.source.time][measure.statistic_journal.statistic.name][measure.quartile] = \
                measure.statistic_journal.value
        data, times = defaultdict(list), []
        for time in sorted(raw_data.keys()):
            times.append(time)
            sub_data = raw_data[time]
            for statistic in statistic_names:
                if statistic in sub_data:
                    data[statistic].append(sub_data[statistic])
                else:
                    data[statistic].append(None)
        print(data)
        print(times)
        return DataFrame(data, index=times)


def data(*args):
    '''
    Start here to access data.  Create an instance in Jupyter:

        d = data('-v', '4')
        print(d.activities())
        ...
    '''
    p = parser()
    args = list(args)
    args.append(NO_OP)
    ns = NamespaceWithVariables(p.parse_args(args))
    log = make_log(ns)
    db = Database(ns, log)
    return Data(log, db)




from hashlib import md5
from os import stat
from shutil import get_terminal_size

from sqlalchemy import desc

from .date import to_time, local_date_to_time
from .schedule import ZERO
from ..squeal.database import add
from ch2.squeal.tables.fit import FileScan


def terminal_width(width=None):
    return get_terminal_size()[0] if width is None else width


def tui(command):
    def wrapper(*args, **kargs):
        return command(*args, **kargs)
    wrapper.tui = True
    wrapper.__doc__ = command.__doc__
    return wrapper


# https://stackoverflow.com/a/3431838
def md5_hash(file_path):
    hash = md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b''):
            hash.update(chunk)
    return hash.hexdigest()


def modified_files(log, s, paths, force=False):
    for file_path in paths:
        last_modified = to_time(stat(file_path).st_mtime)
        hash = md5_hash(file_path)

        path_scan = s.query(FileScan).filter(FileScan.path == file_path).one_or_none()
        if path_scan:
            if hash != path_scan.md5_hash:
                log.warn('File at %s appears to have changed since last read on %s')
                path_scan.md5_hash = hash
                path_scan.last_scan = ZERO
        else:
            path_scan = add(s, FileScan(path=file_path, md5_hash=hash, last_scan=to_time(0.0)))
            s.flush()

        hash_scan = s.query(FileScan).filter(FileScan.md5_hash == hash).\
            order_by(desc(FileScan.last_scan)).limit(1).one()  # must exist as path_scan is a candidate
        if hash_scan.path != path_scan.path:
            log.warn('File at %s appears to be identical to file at %s' % (file_path, hash_scan.path))

        if force or last_modified > hash_scan.last_scan:
            path_scan.last_scan = last_modified
            yield file_path
        else:
            log.debug('Skipping %s (already scanned)' % file_path)


def for_modified_files(log, session, paths, callback, force=False):
    '''
    This takes a callback because we need to know whether to mark the file as read or not
    after processing.  The callback should return True on success.

    The session is used throughout, but not passed to the callback.  The callback can
    contain the same session as internal state, or create its own.  We avoid open
    transactions across the callback.
    '''

    for file_path in paths:

        last_modified = to_time(stat(file_path).st_mtime)
        hash = md5_hash(file_path)

        path_scan = session.query(FileScan).filter(FileScan.path == file_path).one_or_none()
        if path_scan:
            if hash != path_scan.md5_hash:
                log.warn('File at %s appears to have changed since last read on %s')
                path_scan.md5_hash = hash
                path_scan.last_scan = ZERO
        else:
            path_scan = add(session, FileScan(path=file_path, md5_hash=hash, last_scan=0.0))
            session.flush()

        hash_scan = session.query(FileScan).filter(FileScan.md5_hash == hash).\
            order_by(desc(FileScan.last_scan)).limit(1).one()  # must exist as path_scan is a candidate
        if hash_scan.path != path_scan.path:
            log.warn('File at %s appears to be identical to file at %s' % (file_path, hash_scan.path))

        session.commit()

        if force or last_modified > hash_scan.last_scan:
            if callback(file_path):
                log.debug('Marking %s as scanned' % file_path)
                path_scan.last_scan = last_modified
                session.commit()
            else:
                log.debug('Not marking %s as scanned' % file_path)
        else:
            log.debug('Skipping %s (already scanned)' % file_path)




from abc import abstractmethod, ABC

from urwid import Edit, connect_signal

from ch2.lib.utils import label
from ..uweird.tui.widgets import Rating
from ..squeal.tables.statistic import StatisticType

PAGE_WIDTH = 4


class Base(ABC):

    def __init__(self, log, s, journal, width=1):
        self._log = log
        self.__session = s
        self.__journal = journal
        self.width = width

    def __str__(self):
        if self.__journal.value is None:
            return '%s: _' % self.__journal.statistic.name + self._format_units()
        else:
            return '%s: ' % self.__journal.statistic.name + \
                   self._format_value(self.__journal.value) + self._format_units()

    @abstractmethod
    def _format_value(self, value):
        pass

    def _format_units(self):
        return (' ' + self.__journal.statistic.units) if self.__journal.statistic.units else ''

    def bound_widget(self):
        widget = self._widget(self.__journal)
        connect_signal(widget, 'change', self.__on_change)
        return widget

    def __on_change(self, widget, value):
        self._log.debug('Setting %s=%r' % (self.__journal.statistic.name, value))
        self.__journal.value = value

    @abstractmethod
    def _widget(self, journal):
        pass


class Text(Base):

    statistic_type = StatisticType.TEXT

    def __init__(self, log, s, journal, width=PAGE_WIDTH):
        super().__init__(log, s, journal, width=width)

    def _format_value(self, value):
        return repr(value)

    def _widget(self, journal):
        return Edit(caption=label('%s: ' % journal.statistic.name), edit_text=journal.value or '')


class Integer(Base):

    statistic_type = StatisticType.INTEGER

    def __init__(self, log, s, journal, lo=None, hi=None, width=1):
        super().__init__(log, s, journal, width=width)
        self._lo = lo
        self._hi = hi

    def _format_value(self, value):
        return '%d' % value

    def _widget(self, journal):
        from .tui.widgets import Integer
        return Integer(caption=label('%s: ' % journal.statistic.name), state=journal.value,
                       minimum=self._lo, maximum=self._hi, units=journal.statistic.units)


class Float(Base):

    statistic_type = StatisticType.FLOAT

    def __init__(self, log, s, journal, lo=None, hi=None, dp=2, format='%f', width=1):
        super().__init__(log, s, journal, width=width)
        self._lo = lo
        self._hi = hi
        self._dp = dp
        self._format = format

    def _format_value(self, value):
        return self._format % value

    def _widget(self, journal):
        from .tui.widgets import Float
        return Float(caption=label('%s: ' % journal.statistic.name), state=journal.value,
                     minimum=self._lo, maximum=self._hi, dp=self._dp, units=journal.statistic.units)


class Score(Base):

    statistic_type = StatisticType.INTEGER

    def __init__(self, log, s, journal, width=1):
        super().__init__(log, s, journal, width=width)

    def _format_value(self, value):
        return '%d' % value

    def _widget(self, journal):
        return Rating(caption=label('%s: ' % journal.statistic.name), state=journal.value)




from sqlalchemy import Column, Text, Integer, ForeignKey, Float, UniqueConstraint
from sqlalchemy.orm import relationship, backref

from .source import Source, SourceType
from ..support import Base
from ..types import Time, Sort


class Activity(Base):

    __tablename__ = 'activity'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False, server_default='')
    description = Column(Text, nullable=False, server_default='')
    sort = Column(Sort, nullable=False, server_default='')

    @staticmethod
    def lookup(log, s, name):
        activity = s.query(Activity).filter(Activity.name == name).one_or_none()
        if not activity:
            activities = s.query(Activity).all()
            if activities:
                log.info('Available activities:')
                for activity in activities:
                    log.info('%s - %s' % (activity.name, activity.description))
            else:
                log.error('No activities defined - configure system correctly')
            raise Exception('Activity "%s" is not defined' % name)
        return activity


class ActivityJournal(Source):

    __tablename__ = 'activity_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    activity_id = Column(Integer, ForeignKey('activity.id'), nullable=False)
    activity = relationship('Activity')
    name = Column(Text, unique=True)
    fit_file = Column(Text, nullable=False, unique=True)
    finish = Column(Time, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.ACTIVITY
    }

    def __str__(self):
        return 'Activity Journal from %s' % self.time


class ActivityTimespan(Base):

    __tablename__ = 'activity_timespan'

    id = Column(Integer, primary_key=True)
    activity_journal_id = Column(Integer, ForeignKey('source.id', ondelete='cascade'),
                                 nullable=False)
    activity_journal = relationship('ActivityJournal',
                                    backref=backref('timespans', cascade='all, delete-orphan',
                                                    passive_deletes=True,
                                                    order_by='ActivityTimespan.start'))
    start = Column(Time, nullable=False)
    finish = Column(Time, nullable=False)
    UniqueConstraint('activity_journal_id', 'start')


class ActivityWaypoint(Base):

    __tablename__ = 'activity_waypoint'

    activity_journal_id = Column(Integer, ForeignKey('activity_journal.id', ondelete='cascade'),
                                 nullable=False, primary_key=True)
    activity_journal = relationship('ActivityJournal',
                                    backref=backref('waypoints', cascade='all, delete-orphan',
                                                    passive_deletes=True,
                                                    order_by='ActivityWaypoint.time'))
    activity_timespan_id = Column(Integer, ForeignKey('activity_timespan.id'))
    activity_timespan = relationship('ActivityTimespan',
                                     backref=backref('waypoints',
                                                     order_by='ActivityWaypoint.time'))
    time = Column(Time, primary_key=True)
    latitude = Column(Float)
    longitude = Column(Float)
    heart_rate = Column(Integer)
    distance = Column(Float)
    speed = Column(Float)




import datetime as dt
from re import search

from urwid import Text, Pile, Columns, Divider

from . import Displayer
from .heart_rate import build_zones
from ..calculate.activity import ActivityStatistics
from ..names import ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED, MEDIAN_KM_TIME_ANY, MAX_MED_HR_OVER_M_ANY
from ...lib.date import format_seconds, local_date_to_time
from ...lib.utils import label
from ...squeal.tables.activity import Activity, ActivityJournal
from ...squeal.tables.source import Source
from ...squeal.tables.statistic import StatisticJournal, Statistic
from ...uweird.tui.decorators import Indent

HRZ_WIDTH = 30


class ActivityDiary(Displayer):

    def build(self, s, f, date):
        start = local_date_to_time(date)
        finish = start + dt.timedelta(days=1)
        for activity in s.query(Activity).order_by(Activity.sort).all():
            for journal in s.query(ActivityJournal). \
                    filter(ActivityJournal.time >= start,
                           ActivityJournal.time < finish,
                           ActivityJournal.activity == activity). \
                    order_by(ActivityJournal.time).all():
                yield self.__journal_data(s, journal, date)

    def __journal_data(self, s, ajournal, date):
        zones = build_zones(s, ajournal, HRZ_WIDTH)
        details = Pile(self.__active_data(s, ajournal, date))
        return Pile([
            Text(ajournal.name),
            Indent(Columns([details, (HRZ_WIDTH + 2, zones)])),
            Divider(),
            Indent(Columns([Pile(self.__template(s, ajournal, MEDIAN_KM_TIME_ANY,
                                                 'Median Time', r'(\d+km)', date)),
                            Pile(self.__template(s, ajournal, MAX_MED_HR_OVER_M_ANY,
                                                 'Max Med HR', r'(\d+m)', date))]))
        ])

    def __active_data(self, s, ajournal, date):
        body = [Divider()]
        body.append(Text('%s - %s  (%s)' % (ajournal.time.strftime('%H:%M:%S'), ajournal.finish.strftime('%H:%M:%S'),
                                            format_seconds((ajournal.finish - ajournal.time).seconds))))
        for name in (ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED):
            sjournal = self._journal_at_time(s, ajournal.time, name, ActivityStatistics, ajournal.activity.id)
            body.append(Text([label('%s: ' % sjournal.statistic.name)] + self.__format_value(sjournal, date)))
        return body

    def __template(self, s, ajournal, template, title, re, date):
        body = [Text(title)]
        sjournals = s.query(StatisticJournal).join(Statistic, Source). \
            filter(Source.time == ajournal.time,
                   Statistic.name.like(template),
                   Statistic.owner == ActivityStatistics,
                   Statistic.constraint == ajournal.activity.id).order_by(Statistic.name).all()
        # extract
        for sjournal in sorted(sjournals,
                               # order by integer embedded in name
                               key=lambda sjournal: int(search(r'(\d+)', sjournal.statistic.name).group(1))):
            body.append(Text([label(search(re, sjournal.statistic.name).group(1) + ': ')] +
                             self.__format_value(sjournal, date)))
        return body

    def __format_value(self, sjournal, date):
        words, first = ['%s ' % sjournal.formatted()], True
        for measure in sorted(sjournal.measures,
                              key=lambda measure: measure.source.schedule.frame_length_in_days(date)):
            if not first:
                words += [',']
            first = False
            quintile = 1 + min(4, measure.percentile / 20)
            words += [('quintile-%d' % quintile, '%d%%' % int(measure.percentile))]
            if measure.rank < 5:
                words += [':', ('rank-%d' % measure.rank, '%d' % measure.rank)]
            words += ['/' + measure.source.schedule.describe(compact=True)]
        return words




import datetime as dt

from sqlalchemy import or_
from urwid import MainLoop, Columns, Pile, Frame, Filler, Text, Divider, WEIGHT

from .args import DATE
from ..lib.date import to_date, local_date_to_time
from ..lib.io import tui
from ..lib.utils import PALETTE_RAINBOW, em
from ..lib.widgets import DateSwitcher
from ..squeal.tables.pipeline import PipelineType
from ..squeal.tables.source import disable_interval_cleaning, Source
from ..squeal.tables.topic import Topic, TopicJournal
from ..stoats.display import build_pipeline
from ..uweird.fields import PAGE_WIDTH
from ..uweird.tui.decorators import Border, Indent
from ..uweird.tui.factory import Factory
from ..uweird.tui.tabs import TabList
from ..uweird.tui.widgets import DividedPile


@tui
def diary(args, log, db):
    '''
# diary

    ch2 diary [date]

The date can be an absolute day or the number of days previous.  So `ch2 diary 1` selects yesterday.

The daily diary.  Enter information here.

To exit, alt-q (or, without saving, alt-x).
    '''
    date = args[DATE]
    if not date:
        date = dt.date.today()
    else:
        try:
            date = to_date(date)
        except:
            date = dt.date.today() - dt.timedelta(days=int(date))
    disable_interval_cleaning()
    TopicJournal.check_tz(db)
    MainLoop(Diary(log, db, date), palette=PALETTE_RAINBOW).run()


# todo - change to date + tz


class Diary(DateSwitcher):
    '''
    Render the diary at a given date.
    '''

    def _build(self, s):
        self._log.debug('Building diary at %s' % self._date)
        body, f = [], Factory(TabList())
        root_topics = [topic for topic in
                       s.query(Topic).filter(Topic.parent == None,
                                             or_(Topic.start <= self._date, Topic.start == None),
                                             or_(Topic.finish >= self._date, Topic.finish == None)).
                           order_by(Topic.sort).all()
                       if topic.schedule.at_location(self._date)]
        for topic in root_topics:
            body.append(self.__topic(s, f, topic))
        for extra in self.__pipeline(s, f):
            body.append(extra)
        body = Border(Frame(Filler(DividedPile(body), valign='top'),
                            header=Pile([Text(self._date.strftime('%Y-%m-%d - %A')), Divider()]),
                            footer=Pile([Divider(), Text(self.__footer(), align='center')])))
        return body, f.tabs

    def __footer(self):
        footer = ['meta-', em('q'), 'uit/e', em('x'), 'it/', em('s'), 'ave']
        footer += [' [shift]meta-']
        for sep, c in enumerate('dwmya'):
            if sep: footer += ['/']
            footer += [em(c)]
        footer += ['ctivity/', em('t'), 'oday']
        return footer

    def __topic(self, s, f, topic):
        self._log.debug('%s' % topic)
        body, title = [], None
        if topic.name:
            title = Text(topic.name)
        if topic.description:
            body.append(Text(topic.description))
        body += list(self.__fields(s, f, topic))
        body += list(self.__children(s, f, topic))
        if not body:
            return title
        body = Indent(Pile(body))
        if title:
            body = Pile([title, body])
        return body

    def __fields(self, s, f, topic):
        columns, width = [], 0
        for field in topic.fields:
            self._log.debug('%s' % field)
            tjournal = self.__topic_journal(s, topic)
            tjournal.populate(self._log, s)
            display = field.display_cls(self._log, s, tjournal.statistics[field],
                                        *field.display_args, **field.display_kargs)
            self._log.debug('%s' % display)
            if width + display.width > PAGE_WIDTH:
                yield Columns(columns)
                columns, width = [], 0
            columns.append((WEIGHT, display.width, f(display.bound_widget())))
            width += display.width
        if width:
            yield Columns(columns)

    def __children(self, s, f, topic):
        for child in topic.children:
            if child.schedule.at_location(self._date):
                extra = self.__topic(s, f, child)
                if extra:
                    yield extra

    def __topic_journal(self, s, topic):
        tjournal = s.query(TopicJournal). \
            filter(TopicJournal.topic == topic,
                   TopicJournal.date == self._date).one_or_none()
        if not tjournal:
            tjournal = TopicJournal(topic=topic, date=self._date, time=local_date_to_time(self._date))
            s.add(tjournal)
        return tjournal

    def __pipeline(self, s, f):
        yield from build_pipeline(self._log, s, PipelineType.DIARY, f, self._date)

    def save(self):
        if self._session:
            self.__interval_cleaning()
        super().save()

    def __interval_cleaning(self):
        # on exit:
        # - remove any journal entries with no data (all null)
        # - remove any intervals affected by journals with non-null data
        s, dirty = self._session, False
        for tjournal in s.query(TopicJournal).filter(TopicJournal.date == self._date).all():
            clean = True
            tjournal.populate(self._log, s)
            for field in tjournal.topic.fields:
                if tjournal.statistics[field].value is not None:
                    clean = False
                    break
            if clean:
                s.delete(tjournal)
            else:
                dirty = True
        if dirty:
            Source.clean_times(s, [local_date_to_time(self._date)])




from sqlalchemy import Column, Integer, ForeignKey, Text
from sqlalchemy.orm import relationship

from .statistic import Statistic, StatisticJournal
from .source import Source, SourceType
from ..support import Base


class Constant(Base):

    __tablename__ = 'constant'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False)  # StatisticType
    name = Column(Text, nullable=False)
    statistic_id = Column(Integer, ForeignKey('statistic.id', ondelete='cascade'), nullable=False)
    statistic = relationship('Statistic')

    # todo - this called once.  move to caller
    @staticmethod
    def lookup_like(log, s, name):
        constant = s.query(Constant).filter(Constant.name.like(name)).order_by(Constant.name).all()
        if not constant:
            constants = s.query(Constant).all()
            if constants:
                log.info('Available constants:')
                for constant in constants:
                    log.info('%s - %s' % (constant.statistic.name, constant.statistic.description))
            else:
                log.error('No constants defined - configure system correctly')
            raise Exception('Constant "%s" is not defined' % name)
        return constant


class ConstantJournal(Source):

    __tablename__ = 'constant_journal'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)

    @staticmethod
    def lookup_statistic_journal(log, s, name, constraint, time):
        # order important in join here
        # todo - what is join with constant for?  using source as constraint?  should be owner
        # todo - does it even work?
        return s.query(StatisticJournal).join(ConstantJournal, Statistic, Constant). \
            filter(Statistic.constraint == constraint,
                   Statistic.name == name,
                   ConstantJournal.time <= time).one_or_none()

    __mapper_args__ = {
        'polymorphic_identity': SourceType.CONSTANT
    }


class SystemConstant(Base):

    __tablename__ = 'system_constant'

    name = Column(Text, primary_key=True)
    value = Column(Text, nullable=False)




from enum import IntEnum

from sqlalchemy import ForeignKey, Column, Integer, func, and_
from sqlalchemy.event import listens_for, remove
from sqlalchemy.orm import Session, aliased

from ..support import Base
from ..types import Time, OpenSched, Owner
from ...lib.date import to_time
from ...lib.schedule import TZSchedule


class SourceType(IntEnum):

    SOURCE = 0
    INTERVAL = 1
    ACTIVITY = 2
    TOPIC = 3
    CONSTANT = 4
    MONITOR = 5


class Source(Base):

    __tablename__ = 'source'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False)
    time = Column(Time, nullable=False)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.SOURCE,
        'polymorphic_on': type
    }

    @classmethod
    def before_flush(cls, session):
        cls.__clean_dirty_intervals(session)

    @classmethod
    def __clean_dirty_intervals(cls, session):
        times = set()
        for always, instances in [(True, session.new), (False, session.dirty), (True, session.deleted)]:
            # wipe the containing intervals if this is a source that has changed in some way
            # and it's not an interval itself
            sources = [instance for instance in instances
                       if (isinstance(instance, Source) and
                           not isinstance(instance, Interval) and
                           instance.time is not None and
                           (always or session.is_modified(instance)))]
            times |= set(to_time(source.time) for source in sources)
        if times:
            cls.clean_times(session, times)

    @classmethod
    def clean_times(cls, session, times):
        from ...stoats.calculate.summary import SummaryStatistics
        schedules = [TZSchedule(schedule) for schedule in SummaryStatistics.pipeline_schedules(session)]
        for time in times:
            for schedule in schedules:
                start = schedule.start_of_frame_time(time)
                interval = session.query(Interval). \
                    filter(Interval.time == start, Interval.schedule == schedule).one_or_none()
                if interval:
                    session.delete(interval)


@listens_for(Session, 'before_flush')
def before_flush(session, context, instances):
    Source.before_flush(session)


def disable_interval_cleaning():
    remove(Session, 'before_flush', before_flush)


class Interval(Source):

    __tablename__ = 'interval'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    schedule = Column(OpenSched, nullable=False)
    # disambiguate creator so each can wipe only its own data on force
    owner = Column(Owner, nullable=False)
    # duplicates data for simplicity in processing
    # day after (exclusive date)
    finish = Column(Time, nullable=False, index=True)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.INTERVAL
    }

    def __str__(self):
        return 'Interval "%s from %s"' % (self.schedule, self.time)

    @classmethod
    def _missing_interval_starts(cls, log, s, schedule, owner):
        start, finish = cls._raw_statistics_time_range(s)
        finish = schedule.next_frame_time(finish)
        log.debug('Statistics exist %s - %s' % (start, finish))
        starts = cls._open_intervals(s, schedule, owner)
        if not cls._get_interval(s, schedule, owner, start):
            starts = [schedule.start_of_frame_time(start)] + starts
        if starts and starts[-1] == finish:
            starts = starts[:-1]
        log.debug('Have %d open blocks finishing at %s' % (len(starts), finish))
        return starts, finish

    @classmethod
    def _raw_statistics_time_range(cls, s):
        from .statistic import StatisticJournal, Statistic
        start, finish = s.query(func.min(Source.time), func.max(Source.time)). \
            outerjoin(StatisticJournal). \
            filter(Statistic.id != None,
                   Source.time > to_time(0.0)).one()
        if start and finish:
            return start, finish
        else:
            raise Exception('No statistics are currently defined')

    @classmethod
    def _open_intervals(cls, s, schedule, owner):
        close = aliased(Interval)
        return [result[0] for result in s.query(Interval.finish). \
            outerjoin(close,
                      and_(Interval.finish == close.time,
                           Interval.owner == close.owner,
                           Interval.schedule == close.schedule)). \
            filter(close.time == None,
                   Interval.owner == owner,
                   Interval.schedule == schedule). \
            order_by(Interval.finish).all()]

    @classmethod
    def _get_interval(cls, s, schedule, owner, start):
        return s.query(Interval). \
            filter(Interval.time == start,
                   Interval.schedule == schedule,
                   Interval.owner == owner).one_or_none()

    @classmethod
    def _missing_intervals_from(cls, s, schedule, owner, start, finish):
        while start < finish:
            next = schedule.next_frame_time(start)
            yield start, next
            start = next
            if cls._get_interval(s, schedule, owner, start):
                return

    @classmethod
    def missing(cls, log, s, schedule, owner):
        starts, overall_finish = cls._missing_interval_starts(log, s, schedule, owner)
        for block_start in starts:
            yield from cls._missing_intervals_from(s, schedule, owner, block_start, overall_finish)




from urwid import Text, Columns, Pile

from . import Displayer
from ..calculate.monitor import MonitorStatistics
from ..names import STEPS, REST_HR
from ...lib.date import to_date
from ...lib.utils import label
from ...uweird.tui.decorators import Indent


class MonitorDiary(Displayer):

    def build(self, s, f, date):
        date = to_date(date)
        columns = self.__fields(s, date)
        if columns:
            yield Pile([Text('Monitor'),
                        Indent(Columns(columns))])

    def __fields(self, s, date):
        steps = self.__field(s, date, STEPS)
        rest_hr = self.__field(s, date, REST_HR)
        if steps or rest_hr:
            return [steps if steps else Text(''), rest_hr if rest_hr else Text('')]
        else:
            return None

    def __field(self, s, date, name):
        sjournal = self._journal_at_date(s, date, name, MonitorStatistics, None)
        if sjournal:
            return Text([label(name + ': '), sjournal.formatted()])
        else:
            return None




from logging import getLogger, DEBUG, Formatter, INFO, StreamHandler
from logging.handlers import RotatingFileHandler
from os.path import join

from ..command.args import COMMAND, LOGS, PROGNAME, VERBOSITY


def make_log(args, tui=False):

    level_unset = args[VERBOSITY] is None
    level = 4 if level_unset else args[VERBOSITY][0]
    level = 10 * (6 - level)

    file_formatter = Formatter('%(levelname)-8s %(asctime)s: %(message)s')
    name = args[COMMAND] if COMMAND in args else PROGNAME
    path = join(args.dir(LOGS), name + '.log')
    file_handler = RotatingFileHandler(path, maxBytes=1e6, backupCount=10)
    file_handler.setLevel(DEBUG)
    file_handler.setFormatter(file_formatter)

    slog = getLogger('sqlalchemy')
    slog.setLevel(INFO)
    slog.addHandler(file_handler)

    mlog = getLogger('matplotlib')
    mlog.setLevel(INFO)
    mlog.addHandler(file_handler)

    log = getLogger(name)
    log.setLevel(DEBUG)
    log.addHandler(file_handler)

    if not tui or not level_unset:
        stderr_formatter = Formatter('%(levelname)8s: %(message)s')
        stderr_handler = StreamHandler()
        stderr_handler.setLevel(level)
        stderr_handler.setFormatter(stderr_formatter)
        log.addHandler(stderr_handler)
        # slog.addHandler(stderr_handler)
        # mlog.addHandler(stderr_handler)

    return log

